from __future__ import annotations

from dataclasses import dataclass
import logging
import struct

from bleak.backends.device import BLEDevice
from bleak.backends.scanner import AdvertisementData

from aiohomekit.controller.abstract import AbstractDescription
from aiohomekit.model.categories import Categories
from aiohomekit.model.status_flags import StatusFlags

UNPACK_HHBB = struct.Struct("<HHBB").unpack
UNPACK_HH = struct.Struct("<HH").unpack

APPLE_MANUFACTURER_ID = 76
HOMEKIT_ADVERTISEMENT_TYPE = 0x06
HOMEKIT_ENCRYPTED_NOTIFICATION_TYPE = 0x11

logger = logging.getLogger(__name__)


@dataclass
class HomeKitAdvertisement(AbstractDescription):
    setup_hash: bytes
    address: str
    state_num: int

    @classmethod
    def from_manufacturer_data(
        cls, name: str, address: str, manufacturer_data: dict[int, bytes]
    ) -> HomeKitAdvertisement:
        if not (data := manufacturer_data.get(APPLE_MANUFACTURER_ID)):
            raise ValueError("Not an Apple device")

        if data[0] != HOMEKIT_ADVERTISEMENT_TYPE:
            raise ValueError("Not a HomeKit device")

        sf = data[2]
        device_id = ":".join(
            data[3:9].hex()[0 + i : 2 + i] for i in range(0, 12, 2)
        ).lower()
        acid, gsn, cn, cv = UNPACK_HHBB(data[9:15])
        sh = data[15:19]

        return cls(
            name=name,
            id=device_id,
            category=Categories(acid),
            status_flags=StatusFlags(sf),
            config_num=cn,
            state_num=gsn,
            setup_hash=sh,
            address=address,
        )

    @classmethod
    def from_cache(
        cls, address: str, id: str, config_num: int, state_num: int
    ) -> HomeKitAdvertisement:
        """Create a HomeKitAdvertisement from a cache entry."""
        return cls(
            name=address,
            id=id,
            category=Categories(0),
            status_flags=StatusFlags(0),
            config_num=config_num,
            state_num=state_num,
            setup_hash=b"",
            address=address,
        )

    @classmethod
    def from_advertisement(
        cls, device: BLEDevice, advertisement_data: AdvertisementData
    ) -> HomeKitAdvertisement:
        if not (mfr_data := advertisement_data.manufacturer_data):
            raise ValueError("No manufacturer data")

        return cls.from_manufacturer_data(device.name, device.address, mfr_data)


@dataclass
class HomeKitEncryptedNotification:
    name: str
    address: str
    id: str
    advertising_identifier: bytes
    encrypted_payload: bytes

    @classmethod
    def from_manufacturer_data(
        cls, name: str, address: str, manufacturer_data: dict[int, bytes]
    ) -> HomeKitAdvertisement:
        if not (data := manufacturer_data.get(APPLE_MANUFACTURER_ID)):
            raise ValueError("Not an Apple device")

        if data[0] != HOMEKIT_ENCRYPTED_NOTIFICATION_TYPE:
            raise ValueError("Not a HomeKit encrypted notification")

        advertising_identifier = data[2:8]
        device_id = ":".join(
            advertising_identifier.hex()[0 + i : 2 + i] for i in range(0, 12, 2)
        ).lower()
        encrypted_payload = data[8:]

        return cls(
            name=name,
            id=device_id,
            address=address,
            advertising_identifier=advertising_identifier,
            encrypted_payload=encrypted_payload,
        )

    @classmethod
    def from_advertisement(
        cls, device: BLEDevice, advertisement_data: AdvertisementData
    ) -> HomeKitAdvertisement:
        if not (mfr_data := advertisement_data.manufacturer_data):
            raise ValueError("No manufacturer data")

        return cls.from_manufacturer_data(device.name, device.address, mfr_data)



from __future__ import annotations

from functools import cached_property, lru_cache
import logging
from typing import Any
import uuid

from bleak.backends.characteristic import BleakGATTCharacteristic
from bleak.backends.device import BLEDevice
from bleak.exc import BleakError
from bleak_retry_connector import BleakClientWithServiceCache

from .const import HAP_MIN_REQUIRED_MTU

CHAR_DESCRIPTOR_ID = "DC46F0FE-81D2-4616-B5D9-6ABDD796939A"
CHAR_DESCRIPTOR_UUID = uuid.UUID(CHAR_DESCRIPTOR_ID)

logger = logging.getLogger(__name__)
ATT_HEADER_SIZE = 3


class BleakCharacteristicMissing(BleakError):
    """Raised when a characteristic is missing from a service."""


class BleakServiceMissing(BleakError):
    """Raised when a service is missing."""


@lru_cache(maxsize=128, typed=True)
def _determine_fragment_size(
    address: str,
    mtu_size: int,
    additional_overhead_size: int,
    handle: BleakGATTCharacteristic,
) -> int:
    """Determine the fragment size for a characteristic based on the MTU."""
    if max_write_without_response_size := handle.max_write_without_response_size:
        logger.debug(
            "%s: Bleak max_write_without_response_size: %s, mtu_size-3: %s",
            address,
            max_write_without_response_size,
            mtu_size - ATT_HEADER_SIZE,
        )
        fragment_size = max(max_write_without_response_size, mtu_size - ATT_HEADER_SIZE)
    else:
        logger.debug(
            "%s: No bleak obj MTU or max_write_without_response_size, using client.mtu_size-3: %s",
            address,
            mtu_size - ATT_HEADER_SIZE,
        )
        fragment_size = mtu_size - ATT_HEADER_SIZE

    if additional_overhead_size:
        # Secure session means an extra 16 bytes of overhead
        fragment_size -= additional_overhead_size

    logger.debug("%s: Using fragment size: %s", address, fragment_size)

    return fragment_size


class AIOHomeKitBleakClient(BleakClientWithServiceCache):
    """Wrapper for bleak.BleakClient that auto discovers the max mtu."""

    def __init__(self, address_or_ble_device: BLEDevice | str, **kwargs: Any) -> None:
        """Wrap bleak."""
        super().__init__(address_or_ble_device, **kwargs)
        if isinstance(address_or_ble_device, BLEDevice):
            self.__name = address_or_ble_device.name or address_or_ble_device.address
        else:
            self.__name = address_or_ble_device
        self._char_cache: dict[tuple[str, str], BleakGATTCharacteristic] = {}
        self._iid_cache: dict[BleakGATTCharacteristic, int] = {}

    async def get_characteristic(
        self, service_uuid: str, characteristic_uuid: str, iid: int | None = None
    ) -> BleakGATTCharacteristic:
        """Get a characteristic from the cache or the BleakGATTServiceCollection.

        We need to do the linear searching ourselves because the BleakGATTServiceCollection
        calls can raise BleakError if there are more than one matching service or characteristic
        and we want to match the service and characteristic together.
        """
        cache_key = (service_uuid, characteristic_uuid, iid)
        if char := self._char_cache.get(cache_key):
            return char
        service_uuid_lower = service_uuid.lower()
        characteristic_uuid_lower = characteristic_uuid.lower()
        possible_matching_chars: list[BleakGATTCharacteristic] = []
        service_matched = False
        for service in self.services.services.values():
            if service.uuid.lower() == service_uuid_lower:
                service_matched = True
                for char in service.characteristics:
                    if char.uuid.lower() == characteristic_uuid_lower:
                        possible_matching_chars.append(char)

        if len(possible_matching_chars) == 1:
            char = possible_matching_chars[0]
            self._char_cache[cache_key] = char
            return char

        if possible_matching_chars:
            logger.debug(
                "%s: Service %s with characteristics %s is ambiguous",
                self.__name,
                service_uuid,
                characteristic_uuid,
            )
            if not iid:
                raise ValueError(
                    f"{self.__name}: The service {service_uuid} and {characteristic_uuid} "
                    "maps to more than one handle, iid must be provided to disambiguate."
                )
            for possible_matching_char in possible_matching_chars:
                possible_matching_iid = await self.get_characteristic_iid(
                    possible_matching_char
                )
                if iid == possible_matching_iid:
                    self._char_cache[cache_key] = possible_matching_char
                    return possible_matching_char

        if not service_matched:
            available_services = [
                service.uuid for service in self.services.services.values()
            ]
            raise BleakServiceMissing(
                f"{self.__name}: Service {service_uuid} not found, available services: {available_services}"
            )
        available_chars = [char.uuid for char in service.characteristics]
        raise BleakCharacteristicMissing(
            f"{self.__name}: Characteristic {characteristic_uuid} not found, available characteristics: {available_chars}"
        )

    async def get_characteristic_iid(
        self: AIOHomeKitBleakClient, char: BleakGATTCharacteristic
    ) -> int | None:
        """Get the iid of a characteristic."""
        if iid := self._iid_cache.get(char):
            return iid
        iid_handle = char.get_descriptor(CHAR_DESCRIPTOR_UUID)
        if iid_handle is None:
            return None
        value = bytes(await self.read_gatt_descriptor(iid_handle.handle))
        iid = int.from_bytes(value, byteorder="little")
        self._iid_cache[char] = iid
        return iid

    @cached_property
    def mtu_size(self) -> int:
        """Return the mtu size of the client."""
        # Avoid bluezdbus/client.py:561: UserWarning: Using default MTU value.
        # Call _acquire_mtu() or set _mtu_size first to avoid this warning.
        if (
            (backend := getattr(self, "_backend", None))
            and hasattr(backend, "_mtu_size")
            and getattr(backend, "_mtu_size", None) is None
        ):
            return HAP_MIN_REQUIRED_MTU
        return max(super().mtu_size, HAP_MIN_REQUIRED_MTU)

    def determine_fragment_size(
        self,
        additional_overhead_size: int,
        handle: BleakGATTCharacteristic,
    ) -> int:
        """Determine the fragment size for a characteristic based on the MTU."""
        return _determine_fragment_size(
            self.address,
            self.mtu_size,
            additional_overhead_size,
            handle,
        )
